(******************************************************************************)
(*   __               ___     CAVE: Concurrent Algorithm VErifier             *)
(*  /     /\  \    / |        Copyright (c) Viktor Vafeiadis                  *)
(* |     /--\  \  /  |---                                                     *)
(*  \__ /    \  \/   |___     See LICENSE.txt for license.                    *)
(*                                                                            *)
(******************************************************************************)
(** Representation of program commands *)
open Misc
open Parsetree
open Exp
open Pure
open Assertions

(** Kind of branches *) 
type can_goto = Cgo_break of int | Cgo_continue of int | Cgo_return

(** Field assignments *)
type fld_assignments = (exp * component * exp) list

(** Kind of memory orders *)
type read_memory_order =
  | Crmo_nonatomic
  | Crmo_relaxed
  | Crmo_acquire
  | Crmo_seq_cst

type write_memory_order =
  | Cwmo_nonatomic
  | Cwmo_relaxed of (component * component * exp list) option
  | Cwmo_release of (component * component * exp list) option
  | Cwmo_seq_cst of (component * component * exp list) option

type cas_memory_order =
  | Ccmo_seq_cst
  | Ccmo_rel_acq
  | Ccmo_release
  | Ccmo_acquire
  | Ccmo_relaxed

(** Information about function declarations *)
type fun_info =
    { fun_param: Id.t list * Id.t list 
                            (** formal parameters (by reference, by value) *);
      fun_modif: IdSet.t    (** modified variables *);
      fun_pre: cprop     (** precondition *);
      fun_post: cprop    (** postcondition *);
      fun_loc: Location.t   (** location for error messages *) }

(** Canonical statement *)
type can_stmt =
  { can_st_desc: can_st_desc    (** statement descriptor *);
    mutable can_st_lv: IdSet.t  (** set of live variables *);
    can_st_loc: Location.t      (** location for error messages *);
  }

(** Canonical command: sequence of statements *)
and can_cmd = can_stmt list

(** Statement descriptor *)
and can_st_desc =
  | Cst_nondet of can_cmd * can_cmd
  | Cst_kill of IdSet.t
  | Cst_assign of Id.t * exp
  | Cst_fldlookup of component list * Id.t * exp * component * read_memory_order
  | Cst_fldassign of component list * fld_assignments * (string * cprop) list ref * write_memory_order
  | Cst_cas of component list * Id.t * exp * component * exp * exp * cas_memory_order * read_memory_order
  | Cst_new of Id.t * exp
  | Cst_dispose of exp * exp
  | Cst_pfcall of (Id.t option * string * fun_info * Id.t list * exp list) list
  | Cst_fcall2 of IdSet.t * cprop * cprop * string
  | Cst_assume of cprop
  | Cst_assume_exp of exp
  | Cst_assert_exp of exp
  | Cst_interfere of component * string
  | Cst_stabilize of component
  | Cst_action_begin of component * cprop * cprop * cprop * IdSet.t
  | Cst_action_end of component * IdSet.t (** free vars of postcondition *)
  | Cst_loop of can_cmd * cprop option
                             (** infinite loop -- exit with break/return *)
  | Cst_goto of can_goto
  | Cst_comment of string

(** Canonical entailments *)
type can_entailment = 
  cprop * can_cmd * can_cmd * cprop * IdSet.t * Location.t

(** Resource initialization *)
type res_init = 
  | Cri_inv of cprop * Location.t
  | Cri_code of IdSet.t * can_cmd * Location.t

(* -------------------------------------------------------------------------- *)
(** {3 Constructors} *)
(* -------------------------------------------------------------------------- *)

val mk_cmd : can_st_desc -> Location.t -> can_cmd
val mk_assume : exp -> Location.t -> can_cmd
val mk_loop : can_cmd -> cprop option -> Location.t -> IdSet.t -> can_cmd
val mk_nondet : can_cmd -> can_cmd -> Location.t -> can_cmd

(** Make a deep copy of a command (preserve lvs and actions) *)
val cmd_copy : can_cmd -> can_cmd

(** Make a deep copy of a command (clear lvs and actions) *)
val cmd_clear_copy : can_cmd -> can_cmd

(* -------------------------------------------------------------------------- *)
(** {3 Inferred actions annotations} *)
(* -------------------------------------------------------------------------- *)

val actlr_clear : (string * can_entailment) list -> unit 
val actlr_iter : 
  (string * cprop -> unit) -> (string * can_entailment) list -> unit 
val actlr_upd : (cprop -> cprop) -> (string * can_entailment) list -> unit 
val actlr_print : (string * can_entailment) list -> unit 

(* -------------------------------------------------------------------------- *)
(** {3 Linearizability} *)
(* -------------------------------------------------------------------------- *)

val ident_lin_res : Id.t
val linear_fun_pre : cprop
val linear_fun_post : cprop

(** Check that a linearizability specification contains no loops
    and return its effectful & pure execution paths *)
val linear_parse_spec : can_cmd -> can_cmd list * can_cmd list

(** [insert_pure_code pure_code c] copies [c] and insert [pure_code] before
    every atomic_block exit *)
val insert_pure_code : can_cmd -> can_cmd -> can_cmd 

(* -------------------------------------------------------------------------- *)
(** {3 Other operations} *)
(* -------------------------------------------------------------------------- *)

val pp_cmd : Format.formatter -> can_cmd -> unit
val pp_linear_annot : Format.formatter -> can_cmd -> unit

(** [mark_live_vars fv_post c] updates [c] in place by marking its
    live variables at each program point. *)
val mark_live_vars : IdSet.t -> can_cmd -> unit

val insert_kill_dead_vars : can_cmd -> IdSet.t -> can_cmd

val clarify_cmd : can_cmd -> can_cmd

(** Unroll the last iteration of all the loops in [c] *)
(* val unfold_loops : can_cmd -> can_cmd *)

val actlr_merge_linear :
  (cprop -> can_cmd -> 'a) ->
  ('a -> cprop -> cprop) ->
  can_cmd ->
  (string * can_entailment) ->
  (string * can_entailment) list

val left_mover_optimization :
  (string * can_entailment) list -> (string * can_entailment) list

val add_prophecy : can_cmd -> can_cmd 

(** Assumes that variables affected by the substitution are not on LHS of 
    assignments *)
val map_cmd : subst -> can_cmd -> can_cmd

val cmd_fcall_concat : can_cmd -> can_cmd -> can_cmd
