%{
(******************************************************************************)
(*   __               ___     CAVE: Concurrent Algorithm VErifier             *)
(*  /     /\  \    / |        Copyright (c) Viktor Vafeiadis                  *)
(* |     /--\  \  /  |---                                                     *)
(*  \__ /    \  \/   |___     See LICENSE.txt for license.                    *)
(*                                                                            *)
(******************************************************************************)
open Misc
open Parsetree

let mkexp d = { pexp_desc = d; pexp_loc = Location.symbol_loc() }
let mkexp_ghost d = { pexp_desc = d; pexp_loc = Location.none }
let mkstm d = { pstm_desc = d; pstm_loc = Location.symbol_loc() }
let mkstm_ghost d = { pstm_desc = d; pstm_loc = Location.none }

let exp_one = mkexp_ghost (Pexp_num 1)

let mk_ref_params cel loc =
  let check_par pe = match pe.pexp_desc with
    | Pexp_ident _ | Pexp_fld _ -> pe 
    | _ -> 
    raise(Location.Parse_error("Syntax error: Reference parameters must be variables.", loc)) in
(*   let check_par = function
    | {pexp_desc = Pexp_ident i; pexp_loc=l} -> (i,l)
    | {pexp_desc = Pexp_ident i; pexp_loc=l} -> (i,l)
    | _ -> 
    raise(Location.Parse_error("Syntax error: Reference parameters must be variables.", loc)) in *)
  List.map check_par cel
  
let mk_indpred_params cel =
  let check_par = function
    | {pexp_desc = Pexp_ident i; pexp_loc=l} -> (i,l)
    | {pexp_loc = l} -> 
    raise(Location.Parse_error("Syntax error: Node parameters must be variables.", l)) in
  List.map check_par cel

let reflexivity defined_afters myname myparams =
  if List.exists (fun (n,_,_,_) -> n=myname) defined_afters then defined_afters else
  let reflexparams = List.map (fun (x,loc) -> (x^"_",loc)) myparams in
  let myidents = List.map (fun (x,_loc) -> mkexp(Pexp_ident(x))) myparams in
  let reflexidents = List.map (fun (x,_loc) -> mkexp(Pexp_ident(x))) reflexparams in
  let equals = List.map2 (fun x y -> mkexp(Pexp_infix("==",x,y))) myidents reflexidents in
  let assn = match equals with
    | []   -> a_prop_empty
    | h::t -> Aprop_exp (List.fold_left (fun x y -> mkexp(Pexp_infix("&&",x,y))) h t) in
  (myname, reflexparams, assn, Location.none)::defined_afters

let read_of_cas_mo mo = match mo with
  | Pcmo_seq_cst -> Prmo_seq_cst
  | Pcmo_rel_acq -> Prmo_acquire
  | Pcmo_release -> Prmo_relaxed
  | Pcmo_acquire -> Prmo_acquire
  | Pcmo_relaxed -> Prmo_relaxed
 
(* implicitly called when no grammar rules apply *)
let parse_error _ =
  raise(
    Location.Parse_error("Syntax error.",
      match !Location.lexbuf with
	| None -> Location.symbol_loc()
	| Some lexbuf ->
	    (* the Parsing library only updates symbol_end_pos when successfully
	     * reducing a grammar rule, so here we ask the lexer for the current
	     * position directly *)
	    Location.mkloc (Parsing.symbol_start_pos()) lexbuf.Lexing.lex_curr_p))


%} /* declarations */

/* tokens */

%token ACQUIRE
%token ACTION
%token AFTER
%token AMPER
%token AMPERAMPER
%token AS
%token ASSUME
%token ATOMIC
%token ATOMIC_READ
%token ATOMIC_WRITE
%token BARBAR
%token <bool> BOOL
%token BREAK
%token CAS
%token CLASS
%token COLON
%token COMMA
%token COMMENT
%token CONSTRUCTOR
%token CONTINUE
%token DISPOSE
%token DO
%token DOT
%token ELSE
%token EMPTY
%token ENSURES
%token EOF
%token EQUAL
%token ESCROW
%token EXTERN
%token <string> IDENT
%token <string> QIDENT
%token <string> FIDENT
%token IF
%token <string> INFIXOP1
%token <string> INFIXOP2
%token <string> INFIXOP3
%token INLINE
%token INTERFERE
%token INVARIANT
%token LBRACE
%token LBRACKET
%token LET
%token LPAREN
%token MINUSGREATER
%token <int> NAT
%token NEW
%token PAR
%token PROTOCOL
%token PROVER_OPTS
%token POINTSTO
%token RBRACE
%token RBRACKET
%token RELAXED
%token RELEASE
%token REL_ACQ
%token REQUIRES
%token RESOURCE
%token RETURN
%token RPAREN
%token <string> STRING
%token SEMI
%token SEQ_CST
%token STARSTAR
%token STAR
%token STATE
%token THEN
%token TOKEN
%token TREE
%token <string> UNARYOP
%token UNDERSCORE
%token VOID
%token WHEN
%token WHILE
%token WITH

/* precedences (increasing) and associativities for expressions */

%nonassoc bELSE
%nonassoc ELSE AS 
%right BARBAR
%left STARSTAR AMPERAMPER
%left POINTSTO
%left INFIXOP1 EQUAL
%left INFIXOP2
%left INFIXOP3 STAR
%nonassoc UNARYOP
%left MINUSGREATER

/* entry points */

%start program
%type <Parsetree.p_program> program

%% /* rules */

/* entry points */
program:
  | p_items               { List.rev $1 }
;
p_items:
  | /* empty */           { [] }
  | p_items decl          { $2 :: $1 }
  | p_items d_vars        { $2 @ $1 }
;
d_vars:
  | IDENT ids_ne SEMI        { List.rev_map (fun (x,y)-> Pdec_var(x,$1,true,y)) $2 }
  | EXTERN IDENT ids_ne SEMI { List.rev_map (fun (x,y)-> Pdec_var(x,$2,false,y)) $3 }
;
d_option:
  | STRING                   { ("", $1, Location.rhs_loc 1) }
  | IDENT                    { ($1, "",Location.rhs_loc 1) }
  | IDENT EQUAL STRING       { ($1, $3,Location.rhs_loc 1) }
;
d_options: 
  | d_option                    { [$1] }  
  | d_option COMMA d_options    { $1 :: $3 }
;
decl:
  | PROVER_OPTS d_options SEMI          { Pdec_options($2) }
  | COMMENT STRING SEMI                 { Pdec_comment($2) }
  | CLASS IDENT LBRACE d_fields RBRACE  { Pdec_class($2, List.rev $4, Location.rhs_loc 2) }
  | LET IDENT LPAREN ids RPAREN EQUAL LBRACKET assn RBRACKET
      { Pdec_indpred($2,$4,$8,Location.symbol_loc()) }
  | RESOURCE IDENT LBRACE inv constructor_decl interfere_decl actions RBRACE
      { Pdec_resource(component_of_string $2,[],$4,$5,$6,List.rev $7,Location.symbol_loc()) }
  | RESOURCE IDENT LPAREN ids RPAREN LBRACE inv constructor_decl interfere_decl actions RBRACE
      { Pdec_resource(component_of_string $2,$4,$7,$8,$9,List.rev $10,Location.symbol_loc()) }
  | VOID INLINE IDENT LPAREN formals RPAREN top_block
      { Pdec_fun($3,"void",$5,(None,None),$7,true,Location.symbol_loc()) }
  | VOID IDENT LPAREN formals RPAREN top_block
      { Pdec_fun($2,"void",$4,(None,None),$6,false,Location.symbol_loc()) }
  | VOID IDENT LPAREN formals RPAREN fun_specs top_block
      { Pdec_fun($2,"void",$4,$6,$7,false,Location.symbol_loc()) }
  | fun_specs VOID IDENT LPAREN formals RPAREN top_block
      { Pdec_fun($3,"void",$5,$1,$7,false,Location.symbol_loc()) }
  | IDENT IDENT LPAREN formals RPAREN top_block
      { Pdec_fun($2,$1,$4,(None,None),$6,false,Location.symbol_loc()) }
  | IDENT IDENT LPAREN formals RPAREN fun_specs top_block
      { Pdec_fun($2,$1,$4,$6,$7,false,Location.symbol_loc()) }
  | fun_specs IDENT IDENT LPAREN formals RPAREN top_block
      { Pdec_fun($3,$2,$5,$1,$7,false,Location.symbol_loc()) }
  | PROTOCOL IDENT LBRACE protocol_states RBRACE { Pdec_protocol($2,$4,Location.symbol_loc()) }
  | ESCROW IDENT opt_sq_ids REQUIRES assn ENSURES assn { Pdec_escrow($2,$3,$5,$7,Location.symbol_loc()) }
;
opt_sq_ids:
  | /* empty */            { [] }
  | LBRACKET ids RBRACKET  { $2 }
;
protocol_states:
  | /* empty */                    { [] }
  | protocol_state protocol_states { $1::$2 }
;
protocol_state:
  | STATE IDENT opt_sq_ids LPAREN IDENT RPAREN opt_after_assns LBRACE assn RBRACE opt_after_assns
                               { ($2, $3, reflexivity ($7@$11) $2 $3, $5, $9, Location.symbol_loc()) }
;
opt_after_assns:
  | /* empty */                { [] }
  | after_assn opt_after_assns { $1 @ $2 }
;
after_assn:
  | AFTER ids_ne                           { List.map (fun (x,loc) -> (x,[],a_prop_empty,loc)) $2 }
  | AFTER IDENT opt_sq_ids REQUIRES a_exp  { [($2,$3,Aprop_exp $5,Location.symbol_loc())] }
;
d_fields:
  | /* empty */                { [] }
  | d_fields IDENT IDENT SEMI  { ("."^$3,$2,Location.rhs_loc 3)::$1 }
;
fun_specs:
  | REQUIRES assn              { (Some $2, None) }
  | ENSURES assn               { (None, Some $2) }
  | REQUIRES assn ENSURES assn { (Some $2, Some $4) }
;
inv_restr:
  | /* empty */            { None }
  | LBRACKET assn RBRACKET { Some $2 }
;
inv:
  | INVARIANT assn         { Some $2 }
  | inv_restr              { $1 }
;
constructor_decl:
  | /* empty */            { ([], []) }
  | CONSTRUCTOR top_block  { $2 }
;
interfere_decl:
  | /* empty */              { None }
  | INTERFERE LBRACE stmts RBRACE  { Some (List.rev $3) }
;
action_decl:
  | ACTION IDENT LPAREN ids RPAREN
        LBRACKET assn RBRACKET
	LBRACKET assn RBRACKET
      { ($2,$4,a_prop_empty,$7,$10,[],Location.symbol_loc()) }
  | ACTION IDENT LPAREN ids RPAREN
        LBRACKET assn RBRACKET
	LBRACKET assn RBRACKET
        LBRACKET assn RBRACKET
      { ($2,$4,$7,$10,$13,[],Location.symbol_loc()) }
  | ACTION IDENT LPAREN ids RPAREN 
        LBRACKET assn RBRACKET
	LBRACKET assn RBRACKET
        LBRACE stmts RBRACE
      { ($2,$4,a_prop_empty,$7,$10,List.rev $13,Location.symbol_loc()) }
  | ACTION IDENT LPAREN ids RPAREN 
        LBRACKET assn RBRACKET
        LBRACKET assn RBRACKET
	LBRACKET assn RBRACKET
        LBRACE stmts RBRACE
      { ($2,$4,$7,$10,$13,List.rev $16,Location.symbol_loc()) }
;
actions: | { [] } | actions action_decl { $2::$1 } ;
ids: | { [] } | ids_ne { $1 } ;
ids_ne:
  | IDENT            { [($1,Location.rhs_loc 1)] }
  | IDENT COMMA ids_ne { ($1,Location.rhs_loc 1)::$3 }
;
tyids:  { [] } | tyids_ne { $1 } ;
tyids_ne:
  | IDENT IDENT      { [($2,$1,Location.rhs_loc 2)] }
  | IDENT IDENT COMMA tyids_ne  { ($2,$1,Location.rhs_loc 2)::$4 }
;
top_block:
  | LBRACE top_stmts RBRACE     { $2 }
;
top_stmts:
  |                             { ([],[]) }
  | stmt stmts                  { ([], $1 :: List.rev $2) }
  | IDENT ids_ne SEMI top_stmts { (List.map (fun (x,y) -> (x,$1,y)) $2 @ fst $4, snd $4) }
;
stmts: | { [] } | stmts stmt { $2::$1 }
;
fldassigns:
  | exp MINUSGREATER IDENT EQUAL exp { [($1, component_of_string ("."^$3), $5)] }
  | exp MINUSGREATER IDENT EQUAL exp COMMA fldassigns { ($1, component_of_string ("."^$3), $5)::$7 }
;
state_annot_ne:
  | IDENT DOT IDENT                              { ($1, $3, []) }
  | IDENT DOT IDENT LBRACKET a_exps_ne RBRACKET  { ($1, $3, $5) }
;
state_annot:
  | /* empty */                                  { None }
  | state_annot_ne                               { Some $1 }
;
opt_write_mo:
  | /* empty */               { Pwmo_seq_cst None }
  | COMMA state_annot_ne      { Pwmo_seq_cst (Some $2) }
  | COMMA SEQ_CST state_annot { Pwmo_seq_cst ($3) }
  | COMMA RELEASE state_annot { Pwmo_release ($3) }
  | COMMA RELAXED state_annot { Pwmo_relaxed ($3) }
;
stmt:
  | fldassigns SEMI                { mkstm(Pstm_fldassign($1, Pwmo_nonatomic)) }
  | ATOMIC_WRITE LPAREN exp MINUSGREATER IDENT COMMA exp opt_write_mo RPAREN SEMI
                                   { mkstm(Pstm_fldassign([($3, component_of_string ("."^$5), $7)], $8)) }
  | IDENT EQUAL opt_exp SEMI       { mkstm(Pstm_assign($1, $3)) }
  | DISPOSE exp SEMI               { mkstm(Pstm_dispose($2,exp_one)) }
  | DISPOSE LPAREN exp COMMA exp RPAREN SEMI { mkstm(Pstm_dispose($3,$5)) }
  | LBRACE stmts RBRACE            { mkstm(Pstm_block(List.rev $2)) }
  | ASSUME LPAREN exp RPAREN SEMI  { mkstm(Pstm_assume($3)) }
  | INTERFERE IDENT DOT IDENT SEMI { mkstm(Pstm_interfere(component_of_string $2,$4)) }
  | exp SEMI                       { mkstm(Pstm_exp($1)) }
  | PAR LBRACE proc_calls RBRACE   { mkstm(Pstm_parblock(List.rev $3)) }
  | RETURN SEMI                    { mkstm(Pstm_return(None)) }
  | RETURN exp SEMI                { mkstm(Pstm_return(Some $2)) }
  | BREAK SEMI                     { mkstm(Pstm_break) }
  | CONTINUE SEMI                  { mkstm(Pstm_continue) }
  | COMMENT STRING SEMI            { mkstm(Pstm_comment($2)) }
  | IF LPAREN opt_exp RPAREN stmt %prec bELSE  { mkstm(Pstm_if($3, $5, mkstm_ghost(Pstm_block []))) }
  | IF LPAREN opt_exp RPAREN stmt ELSE stmt    { mkstm(Pstm_if($3, $5, $7)) }
  | WHILE LPAREN opt_exp RPAREN inv_restr stmt { mkstm(Pstm_while($5, $3, $6)) }
  | ATOMIC opt_when stmt %prec bELSE           { mkstm(Pstm_atomic($2,$3)) }
  | WITH IDENT opt_when stmt %prec bELSE       { mkstm(Pstm_withres(component_of_string $2,$3,$4,"",[])) }
  | WITH IDENT opt_when stmt AS IDENT LPAREN exps RPAREN SEMI { mkstm(Pstm_withres(component_of_string $2,$3,$4,$6,$8)) }
  | DO stmt AS IDENT DOT IDENT LPAREN exps RPAREN SEMI        { mkstm(Pstm_action($2,component_of_string $4,$6,$8)) }
;
proc_call:
  | IDENT LPAREN actuals RPAREN SEMI              { (None,$1,$3) }
  | IDENT EQUAL IDENT LPAREN actuals RPAREN SEMI  { (Some $1,$3,$5) }
;
proc_calls:
  | proc_call             { [$1] }
  | proc_calls proc_call  { $2::$1 }
;
opt_when:
  | /* empty */            { mkexp(Pexp_infix("==", mkexp_ghost(Pexp_num 0), mkexp_ghost(Pexp_num 0))) }
  | WHEN LPAREN exp RPAREN { $3 }
;
read_mo:
  | SEQ_CST       { Prmo_seq_cst }
  | ACQUIRE       { Prmo_acquire }
  | RELAXED       { Prmo_relaxed }
;
opt_read_mo:
  | /* default */       { Prmo_seq_cst }
  | COMMA read_mo       { $2 }
;
cas_mo: 
  | SEQ_CST       { Pcmo_seq_cst }
  | REL_ACQ       { Pcmo_rel_acq }
  | RELEASE       { Pcmo_release }
  | ACQUIRE       { Pcmo_acquire }
  | RELAXED       { Pcmo_relaxed }
; 
opt_cas_mo:
  | /* default */               { (Pcmo_seq_cst, Prmo_seq_cst) }
  | COMMA cas_mo                { ($2, read_of_cas_mo $2) }
  | COMMA cas_mo COMMA read_mo  { ($2, $4) }
;
exp:
  | IDENT                       { mkexp(Pexp_ident($1)) }
  | NAT                         { mkexp(Pexp_num($1)) }
  | BOOL                        { mkexp(Pexp_bool $1) }
  | LPAREN exp RPAREN           { $2 }
  | LPAREN exp AS IDENT RPAREN  { mkexp(Pexp_cast($2,$4)) }
  | UNARYOP exp                 { mkexp(Pexp_prefix($1, $2)) }
  | INFIXOP2 exp %prec UNARYOP  { mkexp(Pexp_prefix($1, $2)) }
  | exp MINUSGREATER IDENT      { mkexp(Pexp_fld($1, component_of_string ("."^$3), Prmo_nonatomic)) }
  | ATOMIC_READ LPAREN exp MINUSGREATER IDENT opt_read_mo RPAREN
                                { mkexp(Pexp_fld($3, component_of_string ("."^$5), $6)) }
  | CAS LPAREN exp MINUSGREATER IDENT COMMA exp COMMA exp opt_cas_mo RPAREN 
                                { mkexp(Pexp_cas($3, component_of_string ("."^$5), $7, $9, fst $10, snd $10)) }
  | NEW LPAREN RPAREN           { mkexp(Pexp_new ("any",exp_one)) }
  | NEW IDENT LPAREN RPAREN     { mkexp(Pexp_new ($2,exp_one)) }
  | NEW IDENT LPAREN exp RPAREN { mkexp(Pexp_new ($2,$4)) }
  | exp BARBAR exp              { mkexp(Pexp_infix("||", $1, $3)) }
  | exp AMPERAMPER exp          { mkexp(Pexp_infix("&&", $1, $3)) }
  | exp STAR exp                { mkexp(Pexp_infix("*", $1, $3)) }
  | exp INFIXOP1 exp            { mkexp(Pexp_infix($2, $1, $3)) }
  | exp INFIXOP2 exp            { mkexp(Pexp_infix($2, $1, $3)) }
  | exp INFIXOP3 exp            { mkexp(Pexp_infix($2, $1, $3)) }
  | IDENT LPAREN actuals RPAREN { mkexp(Pexp_fcall($1, $3)) }
  | FIDENT LPAREN exps RPAREN   { mkexp(Pexp_fun($1, $3)) }
;
opt_exp:
  | STAR          { None }
  | exp           { Some $1 }
;
exps:
  | /* empty */   { [] }
  | exps_ne       { $1 }
;
exps_ne:
  | exp            { [$1] }
  | exp COMMA exps_ne { $1::$3 }
;
a_exps:
  | /* empty */  { [] }
  | a_exps_ne    { $1 }
;
a_exps_ne:
  | a_exp                 { [$1] }
  | a_exp COMMA a_exps_ne { $1::$3 }
;
formals:
  | tyids               { ([],$1) }
  | tyids SEMI tyids { ($1,$3) }
;
actuals:
  | exps              { ([],$1) }
  | exps SEMI exps    { (mk_ref_params $1 (Location.rhs_loc 1), $3) }
  | AMPER exp               { (mk_ref_params [$2] (Location.rhs_loc 2), []) }
  | AMPER exp COMMA actuals { let (a,b) = $4 in (mk_ref_params [$2] (Location.rhs_loc 2) @ a, b) }
;
assn:
  | LPAREN assn RPAREN           { $2 }
  | EMPTY                        { a_prop_empty }
  | a_exp                        { Aprop_exp $1}
  | a_exp POINTSTO a_fields      { Aprop_node(component_of_string "Node",$1,$3) }
  | a_exp POINTSTO a_exp         { Aprop_node(component_of_string "Node",$1,[(Misc.list_link_tag, $3)]) }
  | IF a_exp THEN assn ELSE assn { Aprop_ifthenelse($2,$4,$6) }
  | assn STARSTAR assn           { Aprop_star($1,$3) }
  | assn BARBAR assn             { Aprop_barbar($1,$3) }
  | IDENT COLON LBRACKET assn RBRACKET      { Aprop_box (component_of_string $1, $4) }
  | IDENT LPAREN a_exps RPAREN              { Aprop_indpred($1,[],$3,Location.symbol_loc()) }
  | IDENT LPAREN a_exps SEMI a_exps RPAREN  { Aprop_indpred($1,mk_indpred_params $3,$5,Location.symbol_loc()) }
  | ATOMIC LBRACKET a_exp MINUSGREATER IDENT COMMA IDENT DOT IDENT                             RBRACKET { Aprop_atomic_fld($3,$5,$7,$9,[]) }
  | ATOMIC LBRACKET a_exp MINUSGREATER IDENT COMMA IDENT DOT IDENT LBRACKET a_exps_ne RBRACKET RBRACKET { Aprop_atomic_fld($3,$5,$7,$9,$11) }
  | ESCROW LBRACKET IDENT                             RBRACKET { Aprop_escrow($3,[]) }
  | ESCROW LBRACKET IDENT LBRACKET a_exps_ne RBRACKET RBRACKET { Aprop_escrow($3,$5) }
  | TOKEN LBRACKET IDENT RBRACKET { Aprop_token(mkexp(Pexp_ident($3))) }
;
a_exp:
  | LPAREN a_exp RPAREN          { $2 }
  | IDENT                        { mkexp(Pexp_ident($1)) }
  | QIDENT                       { mkexp(Pexp_ident($1)) }
  | NAT                          { mkexp(Pexp_num($1)) }
  | BOOL                         { mkexp(Pexp_bool($1)) }
  | FIDENT LPAREN a_exps RPAREN  { mkexp(Pexp_fun($1,$3)) }
  | UNARYOP exp                  { mkexp(Pexp_prefix($1, $2)) }
  | INFIXOP2 a_exp %prec UNARYOP { mkexp(Pexp_prefix($1, $2)) }
  | a_exp AMPERAMPER a_exp       { mkexp(Pexp_infix("&&",$1,$3)) }
  | a_exp INFIXOP1 a_exp         { mkexp(Pexp_infix($2,$1,$3)) }
  | a_exp INFIXOP2 a_exp         { mkexp(Pexp_infix($2,$1,$3)) }
  | a_exp INFIXOP3 a_exp         { mkexp(Pexp_infix($2,$1,$3)) }
  | a_exp STAR a_exp             { mkexp(Pexp_infix("*", $1, $3)) }
;
a_fields:
  | UNDERSCORE   { [] }
  | a_fields_ne  { $1 }
;
a_fields_ne:
  | IDENT COLON a_exp                   { [(component_of_string ("."^$1),$3)] }
  | IDENT COLON a_exp COMMA a_fields_ne { (component_of_string ("."^$1),$3)::$5 }
;

%% (* trailer *)
